import copy
from datetime import datetime
from typing import List

from loguru import logger

from optrabot.broker.optionpricedata import OptionStrikeData
from optrabot.broker.order import (Leg, OptionRight, Order, OrderAction,
                                   OrderType, PriceEffect)
from optrabot.managedtrade import ManagedTrade
from optrabot.optionhelper import OptionHelper
from optrabot.signaldata import SignalData
from optrabot.tradetemplate.processor.templateprocessorbase import \
    TemplateProcessorBase
from optrabot.tradetemplate.templatefactory import IronCondor, Template


class IronCondorProcessor(TemplateProcessorBase):
	def __init__(self, template: Template):
		"""
		Initializes the Iron Condor Processor with the given template
		"""
		super().__init__(template)
		self.credit_legs = None
		self.credit_legs_price = None
		self.adjustment_mode = None  # 'normal' or 'alternate' - tracks which side we're adjusting

	def composeEntryOrder(self, signalData: SignalData = None):
		"""
		Composes the entry order for the Iron Condor template
		"""
		super().composeEntryOrder(signalData)
		iron_condor_template :IronCondor = self._template
		short_strike_put = None
		short_strike_call = None
		premium = iron_condor_template.get_premium()

		# Short Strike Determination
		shortStrikeData = iron_condor_template.getShortStrikeData()
		if not shortStrikeData and not premium:
			raise ValueError('Configuration for Short Strike or premium is missing in template!')

		if iron_condor_template.wing == None:
				raise ValueError('Wing size has not been defined!')

		if shortStrikeData:
			if shortStrikeData.offset:
				logger.debug(f'Using Short Strike Offset: {shortStrikeData.offset}')
				short_strike_put =  OptionHelper.roundToStrikePrice(signalData.close - shortStrikeData.offset)
				short_strike_call =  OptionHelper.roundToStrikePrice(signalData.close + shortStrikeData.offset)
			elif shortStrikeData.delta:
				logger.debug(f'Using Short Strike Delta: {shortStrikeData.delta}')
				try:
					short_strike_put = self.get_short_strike_from_delta("SPX", OptionRight.PUT, shortStrikeData.delta)
					short_strike_call = self.get_short_strike_from_delta("SPX", OptionRight.CALL, shortStrikeData.delta)
				except ValueError as value_error:
					logger.error(f'Error while determining Short Strike by Delta: {value_error}')
					return None

			logger.debug(f'Using Wing Size: {iron_condor_template.wing}')
			long_strike_put = self.get_valid_strike(short_strike_put - iron_condor_template.wing, expiration_date=self._determined_expiration_date, higher=True)
			long_strike_call = self.get_valid_strike(short_strike_call + iron_condor_template.wing, expiration_date=self._determined_expiration_date, higher=False)

			# Now create the entry order with its legs as all required strikes are determined
			legs: List[Leg] = []
			legs.append(Leg(action=OrderAction.SELL, strike=short_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.SELL, strike=short_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.BUY, strike=long_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=self._determined_expiration_date, quantity=1))
			legs.append(Leg(action=OrderAction.BUY, strike=long_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=self._determined_expiration_date, quantity=1))
		elif premium:
			logger.debug(f'Trying to create an Iron Condor with premium: ${premium} and Wing Size: {iron_condor_template.wing}')
			self.credit_legs = None
			self.credit_legs_price = None
			self.adjustment_mode = 'normal'
			legs = self.compose_legs_with_min_premium(iron_condor_template.symbol, self._determined_expiration_date, premium, iron_condor_template.wing)
			if not legs:
				raise ValueError('Unable to determine legs for Iron Condor based on premium and wing size!')

		entryOrder = Order(symbol=self._template.symbol, legs=legs, action=OrderAction.BUY_TO_OPEN, quantity=self._template.amount, type=OrderType.LIMIT)
		return entryOrder
	
	def calculate_position_delta_and_price(self, legs: List[Leg], option_strike_data: OptionStrikeData) -> float:
		"""
		Calculates the total delta of the given legs and the mid price of the legs
		"""
		leg_delta = 0.0
		position_delta = 0.0
		for leg in legs:
			if leg.right == OptionRight.CALL:
				leg_delta = option_strike_data.strikeData[leg.strike].get_call_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getCallMidPrice()
			elif leg.right == OptionRight.PUT:
				leg_delta = option_strike_data.strikeData[leg.strike].get_put_delta() * leg.quantity
				leg.midPrice = option_strike_data.strikeData[leg.strike].getPutMidPrice()
			if leg.action == OrderAction.SELL:
				leg_delta *= -1
			position_delta += leg_delta
		return position_delta
	
	def calculate_position_price(self, legs: List[Leg], open: bool = True) -> float:
		"""
		Calculates the total price of the given legs.
		- Positive position prices indicate a net credit (income)
		- Negative position prices indicate a net debit (cost)
		"""
		total_price = 0.0
		for leg in legs:
			if leg.action == OrderAction.BUY:
				if open:
					total_price -= leg.midPrice * leg.quantity
				else:
					total_price += leg.midPrice * leg.quantity
			elif leg.action == OrderAction.SELL:
				if open:
					total_price += leg.midPrice * leg.quantity
				else:
					total_price -= leg.midPrice * leg.quantity
		return total_price
	
	def compose_legs_with_min_premium(self, symbol: str, expiration_date: datetime.date, min_premium: float, wing: float) -> List[Leg] | None:
		"""
		Composes the Iron Condor legs based on the given minimum premium and the wing size.
		"""
		atm_strike = self.broker_connector.get_atm_strike(self._template.symbol)
		if not atm_strike:
			return
		
		current_short_call_strike = atm_strike
		current_short_put_strike = atm_strike
		option_strike_data = self.broker_connector.get_option_strike_data(symbol, expiration_date)
		strikes = option_strike_data.strikeData.keys()
		next_short_call_strike = OptionHelper.get_next_strike(current_short_call_strike, strikes, True)
		next_short_put_strike = OptionHelper.get_next_strike(current_short_put_strike, strikes, False)
		logger.debug(f'ATM Strike: {atm_strike} - starting with Put Strike: {next_short_put_strike} Call Strike: {next_short_call_strike}')

		starting_legs: List[Leg] = []
		starting_legs.append(Leg(action=OrderAction.SELL, strike=next_short_put_strike, symbol=symbol, right=OptionRight.PUT, expiration=expiration_date, quantity=1))
		starting_legs.append(Leg(action=OrderAction.SELL, strike=next_short_call_strike, symbol=symbol, right=OptionRight.CALL, expiration=expiration_date, quantity=1))
		long_strike_put = self.get_valid_strike(next_short_put_strike - self._template.wing, expiration_date, True)
		long_strike_call = self.get_valid_strike(next_short_call_strike + self._template.wing, expiration_date, False)
		starting_legs.append(Leg(action=OrderAction.BUY, strike=long_strike_call, symbol=self._template.symbol, right=OptionRight.CALL, expiration=expiration_date, quantity=1))
		starting_legs.append(Leg(action=OrderAction.BUY, strike=long_strike_put, symbol=self._template.symbol, right=OptionRight.PUT, expiration=expiration_date, quantity=1))

		entry_order_legs = self.determine_order_legs(starting_legs, strikes, option_strike_data, min_premium)

		return entry_order_legs
	
	def determine_order_legs(self, legs: List[Leg], strikes: List[float], option_strike_data, min_premium: float) -> List[Leg] | None:	
		"""
		Determines the according order legs recursively.
		
		The algorithm iterates through strike combinations trying to find a position that:
		1. Has a delta within the acceptable range (±2)
		2. Has a premium >= min_premium
		
		Strategy:
		- Normal mode: Move the overweight side further OTM to balance delta
		- Alternate mode: Move the other side closer to ATM to balance delta while keeping premium above min
		
		The algorithm stores any valid position (delta in range, premium >= min_premium) and continues
		searching for a position with lower premium but still >= min_premium.
		"""
		position_delta = round(self.calculate_position_delta_and_price(legs, option_strike_data) * 100, 2)
		logger.debug(f'Calculated Position Delta: {position_delta}')
		
		delta_in_range = -2 < position_delta < 2
		
		if delta_in_range:
			# Delta is OK, check if price is sufficient
			new_position_price = self.calculate_position_price(legs, True)
			logger.debug(f'Calculated Position Price: {new_position_price:.2f}. Required min price: {min_premium:.2f}')
			
			if new_position_price >= min_premium:
				# Valid position found - store it as the best candidate so far
				# (closer to min_premium than previous stored position)
				if self.credit_legs is None or new_position_price < self.credit_legs_price:
					self.credit_legs = copy.deepcopy(legs)
					self.credit_legs_price = new_position_price
					logger.debug(f'Stored valid position with premium ${new_position_price:.2f}')
				
				# Try to adjust further OTM to get closer to min_premium
				return self.adjust_legs(position_delta, legs, strikes, option_strike_data, min_premium)
			else:
				# Premium dropped below min_premium - return the last stored valid position
				if self.credit_legs is not None:
					logger.debug(f'Premium ${new_position_price:.2f} dropped below min ${min_premium:.2f}. '
								f'Using stored position with premium ${self.credit_legs_price:.2f}')
					return self.credit_legs
				else:
					# No valid position found at all - this shouldn't happen normally
					logger.warning(f'No valid position with premium >= ${min_premium:.2f} found.')
					return None
		else:
			# Delta out of range, need to adjust
			return self.adjust_legs(position_delta, legs, strikes, option_strike_data, min_premium)

	def adjust_legs(self, position_delta: float, legs: List[Leg], strikes: list[float], option_strike_data, min_premium: float) -> List[Leg] | None:
		"""
		Adjusts the legs to balance delta.
		
		In 'normal' mode: Moves the overweight side further OTM
		In 'alternate' mode: Moves the other side closer to ATM to compensate
		
		If normal adjustment causes delta to go out of range and we have a stored valid position,
		we try alternate mode instead.
		"""
		if self.adjustment_mode == 'normal':
			# Normal mode: Move overweight side further OTM
			result = self._adjust_legs_normal(position_delta, legs, strikes, option_strike_data, min_premium)
			if result is not None:
				return result
			
			# Normal adjustment failed (no more strikes or delta went out of range)
			# If we have a stored valid position, try alternate mode
			if self.credit_legs is not None:
				logger.debug(f'Normal adjustment exhausted. Trying alternate adjustment from stored position.')
				self.adjustment_mode = 'alternate'
				# Start fresh from the stored valid position
				alternate_legs = copy.deepcopy(self.credit_legs)
				# Recalculate delta for stored position
				stored_delta = round(self.calculate_position_delta_and_price(alternate_legs, option_strike_data) * 100, 2)
				return self._adjust_legs_alternate(stored_delta, alternate_legs, strikes, option_strike_data, min_premium)
			else:
				return None
		else:
			# Alternate mode
			return self._adjust_legs_alternate(position_delta, legs, strikes, option_strike_data, min_premium)
	
	def _adjust_legs_normal(self, position_delta: float, legs: List[Leg], strikes: list[float], option_strike_data, min_premium: float) -> List[Leg] | None:
		"""
		Normal adjustment: Move the overweight side further OTM to balance delta.
		"""
		if position_delta > 0:
			# Short Put leg need to be further OTM (lower strike)
			for leg in legs:
				if leg.right == OptionRight.PUT and leg.action == OrderAction.SELL:
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, False)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'[Normal] Adjusted Short Put to Strike: {next_strike}')
						
						long_leg = next((l for l in legs if l.right == OptionRight.PUT and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike - self._template.wing, expiration_date=self._determined_expiration_date, higher=True)
							logger.debug(f'[Normal] Adjusted Long Put to Strike: {long_leg.strike}')
					else:
						logger.debug('No further OTM Put Strike found.')
						return None
		else:
			# Short Call leg need to be further OTM (higher strike)
			for leg in legs:
				if leg.right == OptionRight.CALL and leg.action == OrderAction.SELL:
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, True)
					if next_strike is not None:
						leg.strike = next_strike
						logger.debug(f'[Normal] Adjusted Short Call to Strike: {next_strike}')
						
						long_leg = next((l for l in legs if l.right == OptionRight.CALL and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike + self._template.wing, expiration_date=self._determined_expiration_date, higher=False)
							logger.debug(f'[Normal] Adjusted Long Call to Strike: {long_leg.strike}')
					else:
						logger.debug('No further OTM Call Strike found.')
						return None
		
		return self.determine_order_legs(legs, strikes, option_strike_data, min_premium)
	
	def _adjust_legs_alternate(self, position_delta: float, legs: List[Leg], strikes: list[float], option_strike_data, min_premium: float) -> List[Leg] | None:
		"""
		Alternate adjustment: Move the OTHER side closer to ATM to balance delta.
		This increases premium while adjusting delta.
		
		If delta > 0 (put-heavy): Move Call side closer to ATM (lower call strike)
		If delta < 0 (call-heavy): Move Put side closer to ATM (higher put strike)
		"""
		# First check if current position is valid
		delta_in_range = -2 < position_delta < 2
		if delta_in_range:
			position_price = self.calculate_position_price(legs, True)
			if position_price >= min_premium:
				# Found a valid position in alternate mode
				logger.debug(f'[Alternate] Found valid position with delta {position_delta:.2f} and premium ${position_price:.2f}')
				# Store if better than previous
				if self.credit_legs is None or position_price < self.credit_legs_price:
					self.credit_legs = copy.deepcopy(legs)
					self.credit_legs_price = position_price
				# Continue to see if we can get closer to min_premium
		
		if position_delta > 0:
			# Put-heavy: Move Call side closer to ATM (lower strike = more premium from call side)
			for leg in legs:
				if leg.right == OptionRight.CALL and leg.action == OrderAction.SELL:
					# Move to lower (closer to ATM) call strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, False)  # False = lower strike
					if next_strike is not None:
						# Don't cross over ATM
						atm_strike = self.broker_connector.get_atm_strike(self._template.symbol)
						if next_strike <= atm_strike:
							logger.debug(f'[Alternate] Cannot move Call strike below ATM ({atm_strike})')
							return self.credit_legs
						
						leg.strike = next_strike
						logger.debug(f'[Alternate] Adjusted Short Call closer to ATM: {next_strike}')
						
						long_leg = next((l for l in legs if l.right == OptionRight.CALL and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike + self._template.wing, expiration_date=self._determined_expiration_date, higher=False)
							logger.debug(f'[Alternate] Adjusted Long Call to Strike: {long_leg.strike}')
					else:
						logger.debug('[Alternate] No further ITM Call Strike found.')
						return self.credit_legs
		else:
			# Call-heavy: Move Put side closer to ATM (higher strike = more premium from put side)
			for leg in legs:
				if leg.right == OptionRight.PUT and leg.action == OrderAction.SELL:
					# Move to higher (closer to ATM) put strike
					next_strike = OptionHelper.get_next_strike(leg.strike, strikes, True)  # True = higher strike
					if next_strike is not None:
						# Don't cross over ATM
						atm_strike = self.broker_connector.get_atm_strike(self._template.symbol)
						if next_strike >= atm_strike:
							logger.debug(f'[Alternate] Cannot move Put strike above ATM ({atm_strike})')
							return self.credit_legs
						
						leg.strike = next_strike
						logger.debug(f'[Alternate] Adjusted Short Put closer to ATM: {next_strike}')
						
						long_leg = next((l for l in legs if l.right == OptionRight.PUT and l.action == OrderAction.BUY), None)
						if long_leg is not None:
							long_leg.strike = self.get_valid_strike(next_strike - self._template.wing, expiration_date=self._determined_expiration_date, higher=True)
							logger.debug(f'[Alternate] Adjusted Long Put to Strike: {long_leg.strike}')
					else:
						logger.debug('[Alternate] No further ITM Put Strike found.')
						return self.credit_legs
		
		# Recalculate delta after adjustment
		new_delta = round(self.calculate_position_delta_and_price(legs, option_strike_data) * 100, 2)
		
		# Check if still valid
		if -2 < new_delta < 2:
			position_price = self.calculate_position_price(legs, True)
			if position_price >= min_premium:
				logger.debug(f'[Alternate] Valid position: delta {new_delta:.2f}, premium ${position_price:.2f}')
				if self.credit_legs is None or position_price < self.credit_legs_price:
					self.credit_legs = copy.deepcopy(legs)
					self.credit_legs_price = position_price
				# Continue searching for better position
				return self._adjust_legs_alternate(new_delta, legs, strikes, option_strike_data, min_premium)
			else:
				# Premium dropped below min - return stored position
				logger.debug(f'[Alternate] Premium ${position_price:.2f} dropped below min ${min_premium:.2f}')
				return self.credit_legs
		else:
			# Delta went out of range - return stored position
			logger.debug(f'[Alternate] Delta {new_delta:.2f} out of range')
			return self.credit_legs

	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		super().composeTakeProfitOrder(managedTrade, fillPrice)
		logger.debug('Creating take profit order for template {}', self._template.name)
		takeProfitPrice = self._template.calculateTakeProfitPrice(fillPrice)
		logger.debug(f'Calculated take profit price: {takeProfitPrice}')
		tp_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(tp_order_legs)
		takeProfitOrder = Order(symbol=self._template.symbol, legs=tp_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.LIMIT, price=takeProfitPrice)
		return takeProfitOrder
	
	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		super().composeStopLossOrder(managedTrade, fillPrice)
		logger.debug('Creating stop loss order for template {}', self._template.name)
		stopLossPrice = self._template.calculateStopLossPrice(fillPrice)
		sl_order_legs = copy.deepcopy(managedTrade.entryOrder.legs)
		self.invert_leg_actions(sl_order_legs)
		logger.debug(f'Calculated stop loss price: {stopLossPrice}')
		stopLossOrder = Order(symbol=self._template.symbol, legs=sl_order_legs, action=OrderAction.SELL_TO_CLOSE, quantity=self._template.amount, type=OrderType.STOP, price=stopLossPrice)
		return stopLossOrder


from typing import Any, overload

from typing_extensions import Self

from optrabot.broker.tastytrade.order import InstrumentType
from optrabot.broker.tastytrade.session import Session
from optrabot.broker.tastytrade.utils import TastytradeData


class Pair(TastytradeData):
    """
    Dataclass that represents a specific pair in a pairs watchlist.
    """

    left_action: str
    left_symbol: str
    left_quantity: int
    right_action: str
    right_symbol: str
    right_quantity: int


class PairsWatchlist(TastytradeData):
    """
    Dataclass that represents a pairs watchlist object.
    """

    name: str
    order_index: int
    pairs_equations: list[Pair]

    @overload
    @classmethod
    async def a_get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        name: str | None = None,
    ) -> Self | list[Self]:
        """
        Fetches a list of all Tastytrade public pairs watchlists, or a specific one if
        a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the pairs watchlist to fetch.
        """
        if name:
            data = await session._a_get(f"/pairs-watchlists/{name}")
            return cls(**data)
        data = await session._a_get("/pairs-watchlists")
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        name: str | None = None,
    ) -> Self | list[Self]:
        """
        Fetches a list of all Tastytrade public pairs watchlists, or a specific one if
        a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the pairs watchlist to fetch.
        """
        if name:
            data = session._get(f"/pairs-watchlists/{name}")
            return cls(**data)
        data = session._get("/pairs-watchlists")
        return [cls(**i) for i in data["items"]]


class Watchlist(TastytradeData):
    name: str
    watchlist_entries: list[dict[str, Any]] | None = None
    group_name: str = "default"
    order_index: int = 9999


class PublicWatchlist(Watchlist):
    """
    Dataclass that contains symbols from a public watchlist.
    """

    @overload
    @classmethod
    async def a_get(
        cls, session: Session, *, counts_only: bool = False
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        name: str | None = None,
        *,
        counts_only: bool = False,
    ) -> Self | list[Self]:
        """
        Fetches a list of all Tastytrade public watchlists, or a specific one if
        a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to fetch.
        :param counts_only: whether to only fetch the counts of the watchlists.
        """
        if name:
            data = await session._a_get(f"/public-watchlists/{name}")
            return cls(**data)
        data = await session._a_get(
            "/public-watchlists", params={"counts-only": counts_only}
        )
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session, *, counts_only: bool = False) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        name: str | None = None,
        *,
        counts_only: bool = False,
    ) -> Self | list[Self]:
        """
        Fetches a list of all Tastytrade public watchlists, or a specific one if
        a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to fetch.
        :param counts_only: whether to only fetch the counts of the watchlists.
        """
        if name:
            data = session._get(f"/public-watchlists/{name}")
            return cls(**data)
        data = session._get("/public-watchlists", params={"counts-only": counts_only})
        return [cls(**i) for i in data["items"]]


class PrivateWatchlist(Watchlist):
    """
    Dataclass that contains a private watchlist object, with functions to
    update, publish, modify and remove watchlists.
    """

    @overload
    @classmethod
    async def a_get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        name: str | None = None,
    ) -> Self | list[Self]:
        """
        Fetches the user's private watchlists, or a specific one if a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to fetch.
        """
        if name:
            data = await session._a_get(f"/watchlists/{name}")
            return cls(**data)
        data = await session._a_get("/watchlists")
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, name: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        name: str | None = None,
    ) -> Self | list[Self]:
        """
        Fetches the user's private watchlists, or a specific one if a name is provided.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to fetch.
        """
        if name:
            data = session._get(f"/watchlists/{name}")
            return cls(**data)
        data = session._get("/watchlists")
        return [cls(**i) for i in data["items"]]

    @classmethod
    async def a_remove(cls, session: Session, name: str) -> None:
        """
        Deletes the named private watchlist.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to delete.
        """
        await session._a_delete(f"/watchlists/{name}")

    @classmethod
    def remove(cls, session: Session, name: str) -> None:
        """
        Deletes the named private watchlist.

        :param session: the session to use for the request.
        :param name: the name of the watchlist to delete.
        """
        session._delete(f"/watchlists/{name}")

    async def a_upload(self, session: Session) -> None:
        """
        Creates a private remote watchlist identical to this local one.

        :param session: the session to use for the request.
        """
        await session._a_post("/watchlists", json=self.model_dump(by_alias=True))

    def upload(self, session: Session) -> None:
        """
        Creates a private remote watchlist identical to this local one.

        :param session: the session to use for the request.
        """
        session._post("/watchlists", json=self.model_dump(by_alias=True))

    async def a_update(self, session: Session) -> None:
        """
        Updates the existing private remote watchlist.

        :param session: the session to use for the request.
        """
        await session._a_put(
            f"/watchlists/{self.name}", json=self.model_dump(by_alias=True)
        )

    def update(self, session: Session) -> None:
        """
        Updates the existing private remote watchlist.

        :param session: the session to use for the request.
        """
        session._put(f"/watchlists/{self.name}", json=self.model_dump(by_alias=True))

    def add_symbol(self, symbol: str, instrument_type: InstrumentType) -> None:
        """
        Adds a symbol to the watchlist.
        """
        if self.watchlist_entries is None:
            self.watchlist_entries = []
        self.watchlist_entries.append(
            {"symbol": symbol, "instrument-type": instrument_type}
        )

    def remove_symbol(self, symbol: str, instrument_type: InstrumentType) -> None:
        """
        Removes a symbol from the watchlist.
        """
        if self.watchlist_entries is not None:
            self.watchlist_entries.remove(
                {"symbol": symbol, "instrument-type": instrument_type}
            )



import re
from collections import defaultdict
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Any, overload

from pydantic import field_validator, model_validator
from typing_extensions import Self

from optrabot.broker.tastytrade.order import InstrumentType, TradeableTastytradeData
from optrabot.broker.tastytrade.session import Session
from optrabot.broker.tastytrade.utils import TastytradeData, a_paginate, paginate


class OptionType(str, Enum):
    """
    This is an :class:`~enum.Enum` that contains the valid types of options
    and their abbreviations in the API.
    """

    CALL = "C"
    PUT = "P"


class FutureMonthCode(str, Enum):
    """
    This is an :class:`~enum.Enum` that contains the valid month codes for
    futures.

    This is really here for reference, as the API barely uses these codes.
    """

    JAN = "F"
    FEB = "G"
    MAR = "H"
    APR = "J"
    MAY = "K"
    JUN = "M"
    JUL = "N"
    AUG = "Q"
    SEP = "U"
    OCT = "V"
    NOV = "X"
    DEC = "Z"


class Deliverable(TastytradeData):
    """
    Dataclass representing the deliverable for an option.
    """

    id: int
    root_symbol: str
    deliverable_type: str
    description: str
    amount: Decimal
    percent: str
    symbol: str | None = None
    instrument_type: InstrumentType | None = None


class DestinationVenueSymbol(TastytradeData):
    """
    Dataclass representing a specific destination venue symbol for a
    cryptocurrency.
    """

    id: int
    symbol: str
    destination_venue: str
    routable: bool
    max_quantity_precision: int | None = None
    max_price_precision: int | None = None


class QuantityDecimalPrecision(TastytradeData):
    """
    Dataclass representing the decimal precision (number of places) for an
    instrument.
    """

    instrument_type: InstrumentType
    value: int
    minimum_increment_precision: int
    symbol: str | None = None


class Strike(TastytradeData):
    """
    Dataclass representing a specific strike in an options chain, containing
    the symbols for the call and put options.
    """

    strike_price: Decimal
    call: str
    put: str
    call_streamer_symbol: str
    put_streamer_symbol: str


class TickSize(TastytradeData):
    """
    Dataclass representing the tick size for an instrument.
    """

    value: Decimal
    threshold: Decimal | None = None
    symbol: str | None = None


class NestedOptionChainExpiration(TastytradeData):
    """
    Dataclass representing an expiration in a nested options chain.
    """

    expiration_type: str
    expiration_date: date
    days_to_expiration: int
    settlement_type: str
    strikes: list[Strike]


class NestedFutureOptionChainExpiration(TastytradeData):
    """
    Dataclass representing an expiration in a nested future options chain.
    """

    root_symbol: str
    notional_value: Decimal
    underlying_symbol: str
    strike_factor: Decimal
    days_to_expiration: int
    option_root_symbol: str
    expiration_date: date
    expires_at: datetime
    asset: str
    expiration_type: str
    display_factor: Decimal
    option_contract_symbol: str
    stops_trading_at: datetime
    settlement_type: str
    strikes: list[Strike]
    tick_sizes: list[TickSize]


class NestedFutureOptionFuture(TastytradeData):
    """
    Dataclass representing an underlying future in a nested future options
    chain.
    """

    root_symbol: str
    days_to_expiration: int
    expiration_date: date
    expires_at: datetime
    next_active_month: bool
    symbol: str
    active_month: bool
    stops_trading_at: datetime
    maturity_date: date | None = None

    @field_validator("maturity_date", mode="before")
    @classmethod
    def parse_date_with_utc(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.split(" ")[0]
        return value


class FutureEtfEquivalent(TastytradeData):
    """
    Dataclass that represents the ETF equivalent for a future (aka, the number
    of shares of the ETF that are equivalent to one future, leverage-wise).
    """

    symbol: str
    share_quantity: int


class Roll(TastytradeData):
    """
    Dataclass representing a roll for a future.
    """

    name: str
    active_count: int
    cash_settled: bool
    business_days_offset: int
    first_notice: bool


class Cryptocurrency(TradeableTastytradeData):
    """
    Dataclass that represents a Tastytrade cryptocurrency object. Contains
    information about the cryptocurrency and methods to populate that data
    using cryptocurrency symbol(s).
    """

    id: int
    short_description: str
    description: str
    is_closing_only: bool
    active: bool
    tick_size: Decimal
    destination_venue_symbols: list[DestinationVenueSymbol]
    streamer_symbol: str | None = None

    @overload
    @classmethod
    async def a_get(
        cls, session: Session, symbols: list[str] | None = None
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: str | list[str] | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of cryptocurrency objects from the given symbols,
        or a single cryptocurrency if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the symbol(s) to get the cryptocurrencies for.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = await session._a_get(f"/instruments/cryptocurrencies/{symbol}")
            return cls(**data)
        params = {"symbol[]": symbols} if symbols else None
        data = await session._a_get("/instruments/cryptocurrencies", params=params)
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session, symbols: list[str] | None = None) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        symbols: str | list[str] | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of cryptocurrency objects from the given symbols,
        or a single cryptocurrency if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the symbol(s) to get the cryptocurrencies for.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = session._get(f"/instruments/cryptocurrencies/{symbol}")
            return cls(**data)
        params = {"symbol[]": symbols} if symbols else None
        data = session._get("/instruments/cryptocurrencies", params=params)
        return [cls(**i) for i in data["items"]]


class Equity(TradeableTastytradeData):
    """
    Dataclass that represents a Tastytrade equity object. Contains information
    about the equity and methods to populate that data using equity symbol(s).
    """

    id: int
    is_index: bool
    description: str
    lendability: str
    market_time_instrument_collection: str
    is_closing_only: bool
    is_options_closing_only: bool
    active: bool
    is_illiquid: bool
    is_etf: bool
    streamer_symbol: str
    borrow_rate: Decimal | None = None
    cusip: str | None = None
    short_description: str | None = None
    halted_at: datetime | None = None
    stops_trading_at: datetime | None = None
    is_fractional_quantity_eligible: bool | None = None
    tick_sizes: list[TickSize] | None = None
    listed_market: str | None = None
    option_tick_sizes: list[TickSize] | None = None

    @classmethod
    async def a_get_active_equities(
        cls,
        session: Session,
        per_page: int = 1000,
        page_offset: int | None = 0,
        lendability: str | None = None,
    ) -> list[Self]:
        """
        Returns a list of actively traded Equity objects.

        :param session: the session to use for the request.
        :param per_page: the number of equities to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param lendability:
            the lendability of the equities; e.g. 'Easy To Borrow',
            'Locate Required', 'Preborrow'
        """
        params = {
            "per-page": per_page,
            "page-offset": page_offset,
            "lendability": lendability,
        }
        return await a_paginate(
            session.async_client, cls, "/instruments/equities/active", params
        )

    @classmethod
    def get_active_equities(
        cls,
        session: Session,
        per_page: int = 1000,
        page_offset: int | None = 0,
        lendability: str | None = None,
    ) -> list[Self]:
        """
        Returns a list of actively traded Equity objects.

        :param session: the session to use for the request.
        :param per_page: the number of equities to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param lendability:
            the lendability of the equities; e.g. 'Easy To Borrow',
            'Locate Required', 'Preborrow'
        """
        params = {
            "per-page": per_page,
            "page-offset": page_offset,
            "lendability": lendability,
        }
        return paginate(
            session.sync_client, cls, "/instruments/equities/active", params
        )

    @overload
    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        per_page: int = 250,
        page_offset: int | None = 0,
        lendability: str | None = None,
        is_index: bool | None = None,
        is_etf: bool | None = None,
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: str | list[str],
        per_page: int = 250,
        page_offset: int | None = 0,
        lendability: str | None = None,
        is_index: bool | None = None,
        is_etf: bool | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of Equity objects from the given symbols, or a single
        Equity object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the symbol(s) to get the equities for.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param lendability:
            the lendability of the equities; e.g. 'Easy To Borrow',
            'Locate Required', 'Preborrow'
        :param is_index: whether the equities are indexes.
        :param is_etf: whether the equities are ETFs.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = await session._a_get(f"/instruments/equities/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "lendability": lendability,
            "is-index": is_index,
            "is-etf": is_etf,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return await a_paginate(
            session.async_client, cls, "/instruments/equities", params
        )

    @overload
    @classmethod
    def get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        per_page: int = 250,
        page_offset: int | None = 0,
        lendability: str | None = None,
        is_index: bool | None = None,
        is_etf: bool | None = None,
    ) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        symbols: str | list[str],
        *,
        per_page: int = 250,
        page_offset: int | None = 0,
        lendability: str | None = None,
        is_index: bool | None = None,
        is_etf: bool | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of Equity objects from the given symbols, or a single
        Equity object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the symbol(s) to get the equities for.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param lendability:
            the lendability of the equities; e.g. 'Easy To Borrow',
            'Locate Required', 'Preborrow'
        :param is_index: whether the equities are indexes.
        :param is_etf: whether the equities are ETFs.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = session._get(f"/instruments/equities/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "lendability": lendability,
            "is-index": is_index,
            "is-etf": is_etf,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return paginate(session.sync_client, cls, "/instruments/equities", params)


class Option(TradeableTastytradeData):
    """
    Dataclass that represents a Tastytrade option object. Contains information
    about the option and methods to populate that data using option symbol(s).
    """

    active: bool
    strike_price: Decimal
    root_symbol: str
    underlying_symbol: str
    expiration_date: date
    exercise_style: str
    shares_per_contract: int
    option_type: OptionType
    option_chain_type: str
    expiration_type: str
    settlement_type: str
    stops_trading_at: datetime
    market_time_instrument_collection: str
    days_to_expiration: int
    is_closing_only: bool
    expires_at: datetime | None = None
    streamer_symbol: str = ""
    listed_market: str | None = None
    halted_at: datetime | None = None
    old_security_number: str | None = None

    @model_validator(mode="after")
    def set_streamer_symbol(self) -> Self:
        if self.streamer_symbol == "":
            self._set_streamer_symbol()
        return self

    @overload
    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        active: bool | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
        with_expired: bool | None = None,
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: str | list[str],
        *,
        active: bool | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
        with_expired: bool | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of Option objects from the given symbols, or a single
        Option object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the OCC symbol(s) to get the options for.
        :param active: whether the options are active.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param with_expired: whether to include expired options.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = await session._a_get(f"/instruments/equity-options/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "active": active,
            "with-expired": with_expired,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return await a_paginate(
            session.async_client, cls, "/instruments/equity-options", params
        )

    @overload
    @classmethod
    def get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        per_page: int = 250,
        page_offset: int | None = 0,
        active: bool | None = None,
        with_expired: bool | None = None,
    ) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        symbols: str | list[str],
        *,
        active: bool | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
        with_expired: bool | None = None,
    ) -> Self | list[Self]:
        """
        Returns a list of Option objects from the given symbols, or a single
        Option object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: the OCC symbol(s) to get the options for.
        :param active: whether the options are active.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        :param with_expired: whether to include expired options.
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F")
            data = session._get(f"/instruments/equity-options/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "active": active,
            "with-expired": with_expired,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return paginate(session.sync_client, cls, "/instruments/equity-options", params)

    def _set_streamer_symbol(self) -> None:
        if self.strike_price % 1 == 0:
            strike = "{0:.0f}".format(self.strike_price)
        else:
            strike = "{0:.2f}".format(self.strike_price)
            if strike[-1] == "0":
                strike = strike[:-1]

        exp = self.expiration_date.strftime("%y%m%d")
        self.streamer_symbol = (
            f".{self.underlying_symbol}{exp}{self.option_type.value}{strike}"
        )

    @classmethod
    def streamer_symbol_to_occ(cls, streamer_symbol: str) -> str:
        """
        Returns the OCC 2010 symbol equivalent to the given streamer symbol.

        :param streamer_symbol: the streamer symbol to convert
        """
        match = re.match(r"\.([A-Z]+)(\d{6})([CP])(\d+)(\.(\d+))?", streamer_symbol)
        if match is None:
            return ""
        symbol = match.group(1)[:6].ljust(6)
        exp = match.group(2)
        option_type = match.group(3)
        strike = match.group(4).zfill(5)
        if match.group(6) is not None:
            decimal = str(100 * int(match.group(6))).zfill(3)
        else:
            decimal = "000"

        return f"{symbol}{exp}{option_type}{strike}{decimal}"

    @classmethod
    def occ_to_streamer_symbol(cls, occ: str) -> str:
        """
        Returns the dxfeed symbol for use in the streamer from the given OCC
        2010 symbol.

        :param occ: the OCC symbol to convert
        """
        symbol = occ[:6].split()[0]
        info = occ[6:]
        match = re.match(r"(\d{6})([CP])(\d{5})(\d{3})", info)
        if match is None:
            return ""
        exp = match.group(1)
        option_type = match.group(2)
        strike = int(match.group(3))
        decimal = int(match.group(4))

        res = f".{symbol}{exp}{option_type}{strike}"
        if decimal != 0:
            decimal_str = str(decimal / 1000.0)
            res += decimal_str[1:]
        return res


class NestedOptionChain(TastytradeData):
    """
    Dataclass that represents a Tastytrade nested option chain object.
    Contains information about the option chain and a method to fetch one for
    a symbol.

    This is cleaner than calling :meth:`get_option_chain` but if you want to
    create actual :class:`Option` objects you'll need to make an extra API
    request or two.
    """

    underlying_symbol: str
    root_symbol: str
    option_chain_type: str
    shares_per_contract: int
    tick_sizes: list[TickSize]
    expirations: list[NestedOptionChainExpiration]
    deliverables: list[Deliverable] | None = None

    @classmethod
    async def a_get(cls, session: Session, symbol: str) -> list[Self]:
        """
        Gets the option chain for the given symbol in nested format.

        :param session: the session to use for the request.
        :param symbol: the symbol to get the option chain for.
        """
        symbol = symbol.replace("/", "%2F")
        data = await session._a_get(f"/option-chains/{symbol}/nested")
        return [cls(**item) for item in data["items"]]

    @classmethod
    def get(cls, session: Session, symbol: str) -> list[Self]:
        """
        Gets the option chain for the given symbol in nested format.

        :param session: the session to use for the request.
        :param symbol: the symbol to get the option chain for.
        """
        symbol = symbol.replace("/", "%2F")
        data = session._get(f"/option-chains/{symbol}/nested")
        return [cls(**item) for item in data["items"]]


class FutureProduct(TastytradeData):
    """
    Dataclass that represents a Tastytrade future product object. Contains
    information about the future product and a method to fetch one for a
    symbol.

    Useful for fetching general information about a family of futures, without
    knowing the specific expirations or symbols.
    """

    root_symbol: str
    code: str
    description: str
    exchange: str
    product_type: str
    listed_months: list[FutureMonthCode]
    active_months: list[FutureMonthCode]
    notional_multiplier: Decimal
    tick_size: Decimal
    display_factor: Decimal
    streamer_exchange_code: str
    small_notional: bool
    back_month_first_calendar_symbol: bool
    first_notice: bool
    cash_settled: bool
    market_sector: str
    clearing_code: str
    clearing_exchange_code: str
    roll: Roll
    base_tick: int | None = None
    sub_tick: int | None = None
    contract_limit: int | None = None
    product_subtype: str | None = None
    security_group: str | None = None
    true_underlying_code: str | None = None
    clearport_code: str | None = None
    legacy_code: str | None = None
    legacy_exchange_code: str | None = None
    option_products: list["FutureOptionProduct"] | None = None

    @overload
    @classmethod
    async def a_get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(
        cls, session: Session, code: str, exchange: str = "CME"
    ) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        code: str | None = None,
        exchange: str = "CME",
    ) -> Self | list[Self]:
        """
        Returns a list of FutureProduct objects available, or a single
        FutureProduct object if a code is provided.

        :param session: the session to use for the request.
        :param code: the product code, e.g. 'ES'
        :param exchange:
            the exchange to fetch from: 'CME', 'SMALLS', 'CFE', 'CBOED'
        """
        if code:
            code = code.replace("/", "")
            data = await session._a_get(
                f"/instruments/future-products/{exchange}/{code}"
            )
            return cls(**data)
        data = await session._a_get("/instruments/future-products")
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, code: str, exchange: str = "CME") -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        code: str | None = None,
        exchange: str = "CME",
    ) -> Self | list[Self]:
        """
        Returns a list of FutureProduct objects available, or a single
        FutureProduct object if a code is provided.

        :param session: the session to use for the request.
        :param code: the product code, e.g. 'ES'
        :param exchange:
            the exchange to fetch from: 'CME', 'SMALLS', 'CFE', 'CBOED'
        """
        if code:
            code = code.replace("/", "")
            data = session._get(f"/instruments/future-products/{exchange}/{code}")
            return cls(**data)
        data = session._get("/instruments/future-products")
        return [cls(**i) for i in data["items"]]


class Future(TradeableTastytradeData):
    """
    Dataclass that represents a Tastytrade future object. Contains information
    about the future and methods to fetch futures for symbol(s).
    """

    product_code: str
    tick_size: Decimal
    notional_multiplier: Decimal
    display_factor: Decimal
    last_trade_date: date
    expiration_date: date
    active: bool
    active_month: bool
    next_active_month: bool
    is_closing_only: bool
    stops_trading_at: datetime
    expires_at: datetime
    product_group: str
    exchange: str
    streamer_exchange_code: str
    back_month_first_calendar_symbol: bool
    instrument_type: InstrumentType = InstrumentType.FUTURE
    streamer_symbol: str = ""
    closing_only_date: date | None = None
    is_tradeable: bool | None = None
    future_product: "FutureProduct | None" = None
    contract_size: Decimal | None = None
    main_fraction: Decimal | None = None
    sub_fraction: Decimal | None = None
    first_notice_date: date | None = None
    roll_target_symbol: str | None = None
    true_underlying_symbol: str | None = None
    future_etf_equivalent: FutureEtfEquivalent | None = None
    tick_sizes: list[TickSize] | None = None
    option_tick_sizes: list[TickSize] | None = None
    spread_tick_sizes: list[TickSize] | None = None

    @overload
    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: list[str] | None = None,
        *,
        product_codes: list[str] | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: str | list[str] | None = None,
        *,
        product_codes: list[str] | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> Self | list[Self]:
        """
        Returns a list of Future objects from the given symbols
        or product codes.

        :param session: the session to use for the request.
        :param symbols:
            symbol(s) of the futures, e.g. 'ESZ9', '/ESZ9'.
        :param product_codes:
            the product codes of the futures, e.g. 'ES', '6A'. Ignored if
            symbols are provided.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "")
            data = await session._a_get(f"/instruments/futures/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "product-code[]": product_codes,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return await a_paginate(
            session.async_client, cls, "/instruments/futures", params
        )

    @overload
    @classmethod
    def get(
        cls,
        session: Session,
        symbols: list[str] | None = None,
        *,
        product_codes: list[str] | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        symbols: str | list[str] | None = None,
        *,
        product_codes: list[str] | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> Self | list[Self]:
        """
        Returns a list of Future objects from the given symbols
        or product codes.

        :param session: the session to use for the request.
        :param symbols:
            symbol(s) of the futures, e.g. 'ESZ9', '/ESZ9'.
        :param product_codes:
            the product codes of the futures, e.g. 'ES', '6A'. Ignored if
            symbols are provided.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "")
            data = session._get(f"/instruments/futures/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "product-code[]": product_codes,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return paginate(session.sync_client, cls, "/instruments/futures", params)


class FutureOptionProduct(TastytradeData):
    """
    Dataclass that represents a Tastytrade future option product object.
    Contains information about the future option product (deliverable for
    the future option).
    """

    root_symbol: str
    cash_settled: bool
    code: str
    display_factor: Decimal
    exchange: str
    product_type: str
    expiration_type: str
    settlement_delay_days: int
    market_sector: str
    clearing_code: str
    clearing_exchange_code: str
    clearing_price_multiplier: Decimal
    is_rollover: bool
    future_product: "FutureProduct | None" = None
    product_subtype: str | None = None
    legacy_code: str | None = None
    clearport_code: str | None = None

    @overload
    @classmethod
    async def a_get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(
        cls, session: Session, root_symbol: str, exchange: str = "CME"
    ) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        root_symbol: str | None = None,
        exchange: str = "CME",
    ) -> Self | list[Self]:
        """
        Returns a list of FutureOptionProduct objects available, or a single
        FutureOptionProduct object if a root symbol is provided.

        :param session: the session to use for the request.
        :param root_symbol: the root symbol of the future option
        :param exchange: the exchange to get the product from
        """
        if root_symbol:
            root_symbol = root_symbol.replace("/", "")
            data = await session._a_get(
                f"/instruments/future-option-products/{exchange}/{root_symbol}"
            )
            return cls(**data)
        data = await session._a_get("/instruments/future-option-products")
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, root_symbol: str, exchange: str = "CME") -> Self: ...

    @classmethod
    def get(
        cls, session: Session, root_symbol: str | None = None, exchange: str = "CME"
    ) -> Self | list[Self]:
        """
        Returns a list of FutureOptionProduct objects available, or a single
        FutureOptionProduct object if a root symbol is provided.

        :param session: the session to use for the request.
        :param root_symbol: the root symbol of the future option
        :param exchange: the exchange to get the product from
        """
        if root_symbol:
            root_symbol = root_symbol.replace("/", "")
            data = session._get(
                f"/instruments/future-option-products/{exchange}/{root_symbol}"
            )
            return cls(**data)
        data = session._get("/instruments/future-option-products")
        return [cls(**i) for i in data["items"]]


class FutureOption(TradeableTastytradeData):
    """
    Dataclass that represents a Tastytrade future option object. Contains
    information about the future option, and methods to get future options.
    """

    underlying_symbol: str
    product_code: str
    expiration_date: date
    root_symbol: str
    option_root_symbol: str
    strike_price: Decimal
    exchange: str
    streamer_symbol: str
    option_type: OptionType
    exercise_style: str
    is_vanilla: bool
    is_primary_deliverable: bool
    future_price_ratio: Decimal
    multiplier: Decimal
    underlying_count: Decimal
    is_confirmed: bool
    notional_value: Decimal
    display_factor: Decimal
    settlement_type: str
    strike_factor: Decimal
    maturity_date: date
    is_exercisable_weekly: bool
    last_trade_time: str
    days_to_expiration: int
    is_closing_only: bool
    active: bool
    stops_trading_at: datetime
    expires_at: datetime
    exchange_symbol: str
    security_exchange: str
    sx_id: str
    instrument_type: InstrumentType = InstrumentType.FUTURE_OPTION
    future_option_product: "FutureOptionProduct | None" = None

    @field_validator("maturity_date", mode="before")
    @classmethod
    def parse_date_with_utc(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.split(" ")[0]
        return value

    @overload
    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        root_symbol: str | None = None,
        expiration_date: date | None = None,
        option_type: OptionType | None = None,
        strike_price: Decimal | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls,
        session: Session,
        symbols: str | list[str],
        *,
        root_symbol: str | None = None,
        expiration_date: date | None = None,
        option_type: OptionType | None = None,
        strike_price: Decimal | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> Self | list[Self]:
        """
        Returns a list of FutureOption objects from the given symbols.

        NOTE: many of the parameters are bugged, maybe Tasty will fix?

        :param session: the session to use for the request.
        :param symbols: the Tastytrade symbol(s) to filter by.
        :param root_symbol:
            the root symbol to get the future options for, e.g. 'EW3', 'SO'
        :param expiration_date: the expiration date for the future options.
        :param option_type: the option type to filter by.
        :param strike_price: the strike price to filter by.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F").replace(" ", "%20")
            data = await session._a_get(f"/instruments/future-options/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "option-root-symbol": root_symbol,
            "expiration-date": expiration_date,
            "option-type": option_type.value if option_type else None,
            "strike-price": strike_price,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return await a_paginate(
            session.async_client, cls, "/instruments/future-options", params
        )

    @overload
    @classmethod
    def get(
        cls,
        session: Session,
        symbols: list[str],
        *,
        root_symbol: str | None = None,
        expiration_date: date | None = None,
        option_type: OptionType | None = None,
        strike_price: Decimal | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls,
        session: Session,
        symbols: str | list[str],
        *,
        root_symbol: str | None = None,
        expiration_date: date | None = None,
        option_type: OptionType | None = None,
        strike_price: Decimal | None = None,
        per_page: int = 250,
        page_offset: int | None = 0,
    ) -> Self | list[Self]:
        """
        Returns a list of FutureOption objects from the given symbols.

        NOTE: many of the parameters are bugged, maybe Tasty will fix?

        :param session: the session to use for the request.
        :param symbols: the Tastytrade symbol(s) to filter by.
        :param root_symbol:
            the root symbol to get the future options for, e.g. 'EW3', 'SO'
        :param expiration_date: the expiration date for the future options.
        :param option_type: the option type to filter by.
        :param strike_price: the strike price to filter by.
        :param per_page: the number of options to get per page.
        :param page_offset:
            provide a specific page to get; if None, get all pages
        """
        if isinstance(symbols, str):
            symbol = symbols.replace("/", "%2F").replace(" ", "%20")
            data = session._get(f"/instruments/future-options/{symbol}")
            return cls(**data)
        params = {
            "symbol[]": symbols,
            "option-root-symbol": root_symbol,
            "expiration-date": expiration_date,
            "option-type": option_type.value if option_type else None,
            "strike-price": strike_price,
            "per-page": per_page,
            "page-offset": page_offset,
        }
        return paginate(session.sync_client, cls, "/instruments/future-options", params)


class NestedFutureOptionSubchain(TastytradeData):
    """
    Dataclass that represents a Tastytrade nested future option chain for a
    specific futures underlying symbol.
    """

    underlying_symbol: str
    root_symbol: str
    exercise_style: str
    expirations: list[NestedFutureOptionChainExpiration]


class NestedFutureOptionChain(TastytradeData):
    """
    Dataclass that represents a Tastytrade nested option chain object. Contains
    information about the option chain and a method to fetch one for a symbol.

    This is cleaner than calling :meth:`get_future_option_chain` but if you
    want to create actual :class:`FutureOption` objects you'll need to make an
    extra API request or two.
    """

    futures: list[NestedFutureOptionFuture]
    option_chains: list[NestedFutureOptionSubchain]

    @classmethod
    async def a_get(cls, session: Session, symbol: str) -> Self:
        """
        Gets the futures option chain for the given symbol in nested format.

        :param session: the session to use for the request.
        :param symbol: the symbol to get the option chain for.
        """
        symbol = symbol.replace("/", "")
        data = await session._a_get(f"/futures-option-chains/{symbol}/nested")
        return cls(**data)

    @classmethod
    def get(cls, session: Session, symbol: str) -> Self:
        """
        Gets the futures option chain for the given symbol in nested format.

        :param session: the session to use for the request.
        :param symbol: the symbol to get the option chain for.
        """
        symbol = symbol.replace("/", "")
        data = session._get(f"/futures-option-chains/{symbol}/nested")
        return cls(**data)


class Warrant(TastytradeData):
    """
    Dataclass that represents a Tastytrade warrant object. Contains
    information about the warrant, and methods to get warrants.
    """

    symbol: str
    instrument_type: InstrumentType
    listed_market: str
    description: str
    is_closing_only: bool
    active: bool
    cusip: str | None = None

    @overload
    @classmethod
    async def a_get(
        cls, session: Session, symbols: list[str] | None = None
    ) -> list[Self]: ...

    @overload
    @classmethod
    async def a_get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    async def a_get(
        cls, session: Session, symbols: str | list[str] | None = None
    ) -> Self | list[Self]:
        """
        Returns a list of Warrant objects from the given symbols, or a single
        Warrant object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: symbol(s) of the warrants, e.g. 'NKLAW'
        """
        if isinstance(symbols, str):
            data = await session._a_get(f"/instruments/warrants/{symbols}")
            return cls(**data)
        params = {"symbol[]": symbols} if symbols else None
        data = await session._a_get("/instruments/warrants", params=params)
        return [cls(**i) for i in data["items"]]

    @overload
    @classmethod
    def get(cls, session: Session, symbols: list[str] | None = None) -> list[Self]: ...

    @overload
    @classmethod
    def get(cls, session: Session, symbols: str) -> Self: ...

    @classmethod
    def get(
        cls, session: Session, symbols: str | list[str] | None = None
    ) -> Self | list[Self]:
        """
        Returns a list of Warrant objects from the given symbols, or a single
        Warrant object if a list is not provided.

        :param session: the session to use for the request.
        :param symbols: symbol(s) of the warrants, e.g. 'NKLAW'
        """
        if isinstance(symbols, str):
            data = session._get(f"/instruments/warrants/{symbols}")
            return cls(**data)
        params = {"symbol[]": symbols} if symbols else None
        data = session._get("/instruments/warrants", params=params)
        return [cls(**i) for i in data["items"]]


# fix pydantic forward references
FutureProduct.model_rebuild()


async def a_get_quantity_decimal_precisions(
    session: Session,
) -> list[QuantityDecimalPrecision]:
    """
    Returns a list of QuantityDecimalPrecision objects for different
    types of instruments.

    :param session: the session to use for the request.
    """
    data = await session._a_get("/instruments/quantity-decimal-precisions")
    return [QuantityDecimalPrecision(**i) for i in data["items"]]


def get_quantity_decimal_precisions(session: Session) -> list[QuantityDecimalPrecision]:
    """
    Returns a list of QuantityDecimalPrecision objects for different
    types of instruments.

    :param session: the session to use for the request.
    """
    data = session._get("/instruments/quantity-decimal-precisions")
    return [QuantityDecimalPrecision(**i) for i in data["items"]]


async def a_get_option_chain(session: Session, symbol: str) -> dict[date, list[Option]]:
    """
    Returns a mapping of expiration date to a list of option objects
    representing the options chain for the given symbol.

    In the case that there are two expiries on the same day (e.g. SPXW
    and SPX AM options), both will be returned in the same list. If you
    just want one expiry, you'll need to filter the list yourself, or use
    :class:`NestedOptionChain` instead.

    :param session: the session to use for the request.
    :param symbol: the symbol to get the option chain for.
    """
    symbol = symbol.replace("/", "%2F")
    data = await session._a_get(f"/option-chains/{symbol}")
    chain: dict[date, list[Option]] = defaultdict(list)
    for i in data["items"]:
        option = Option(**i)
        chain[option.expiration_date].append(option)

    return chain


def get_option_chain(session: Session, symbol: str) -> dict[date, list[Option]]:
    """
    Returns a mapping of expiration date to a list of option objects
    representing the options chain for the given symbol.

    In the case that there are two expiries on the same day (e.g. SPXW
    and SPX AM options), both will be returned in the same list. If you
    just want one expiry, you'll need to filter the list yourself, or use
    :class:`NestedOptionChain` instead.

    :param session: the session to use for the request.
    :param symbol: the symbol to get the option chain for.
    """
    symbol = symbol.replace("/", "%2F")
    data = session._get(f"/option-chains/{symbol}")
    chain: dict[date, list[Option]] = defaultdict(list)
    for i in data["items"]:
        option = Option(**i)
        chain[option.expiration_date].append(option)

    return chain


async def a_get_future_option_chain(
    session: Session, symbol: str
) -> dict[date, list[FutureOption]]:
    """
    Returns a mapping of expiration date to a list of futures options
    objects representing the options chain for the given symbol.

    In the case that there are two expiries on the same day (e.g. EW
    and ES options), both will be returned in the same list. If you
    just want one expiry, you'll need to filter the list yourself, or
    use :class:`NestedFutureOptionChain` instead.

    :param session: the session to use for the request.
    :param symbol: the symbol to get the option chain for.
    """
    symbol = symbol.replace("/", "")
    data = await session._a_get(f"/futures-option-chains/{symbol}")
    chain: dict[date, list[FutureOption]] = defaultdict(list)
    for i in data["items"]:
        option = FutureOption(**i)
        chain[option.expiration_date].append(option)

    return chain


def get_future_option_chain(
    session: Session, symbol: str
) -> dict[date, list[FutureOption]]:
    """
    Returns a mapping of expiration date to a list of futures options
    objects representing the options chain for the given symbol.

    In the case that there are two expiries on the same day (e.g. EW
    and ES options), both will be returned in the same list. If you
    just want one expiry, you'll need to filter the list yourself, or
    use :class:`NestedFutureOptionChain` instead.

    :param session: the session to use for the request.
    :param symbol: the symbol to get the option chain for.
    """
    symbol = symbol.replace("/", "")
    data = session._get(f"/futures-option-chains/{symbol}")
    chain: dict[date, list[FutureOption]] = defaultdict(list)
    for i in data["items"]:
        option = FutureOption(**i)
        chain[option.expiration_date].append(option)

    return chain



from decimal import Decimal

from .event import IndexedEvent


class Greeks(IndexedEvent):
    """
    Greek ratios, or simply Greeks, are differential values that show how the
    price of an option depends on other market parameters: on the price of the
    underlying asset, its volatility, etc. Greeks are used to assess the risks
    of customer portfolios. Greeks are derivatives of the value of securities
    in different axes. If a derivative is very far from zero, then the
    portfolio has a risky sensitivity in this parameter.
    """

    #: unique per-symbol index of this event
    index: int
    #: timestamp of this event in milliseconds
    time: int
    #: sequence number to distinguish events that have the same time
    sequence: int
    #: option market price
    price: Decimal
    #: Black-Scholes implied volatility of the option
    volatility: Decimal
    #: option delta
    delta: Decimal
    #: option gamma
    gamma: Decimal
    #: option theta
    theta: Decimal
    #: option rho
    rho: Decimal
    #: option vega
    vega: Decimal



"""
Base utilities and data classes for the Tradier API client.
"""

from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Any, TypeVar

from pydantic import BaseModel, ConfigDict

from optrabot.broker.tradier.exceptions import TradierAPIError

if TYPE_CHECKING:
    from httpx import Response


class TradierData(BaseModel):
    """
    Base class for all Tradier data models.
    
    Uses Pydantic's BaseModel with custom configuration for
    flexible field handling and alias support.
    """

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        extra="ignore",
    )


T = TypeVar("T", bound=TradierData)


def validate_response(response: Response) -> None:
    """
    Validate an HTTP response and raise an exception if it indicates an error.
    
    :param response: The httpx Response object to validate.
    :raises TradierAPIError: If the response indicates an error.
    """
    if response.status_code // 100 != 2:
        try:
            error_data = response.json()
            if "fault" in error_data:
                fault = error_data["fault"]
                message = fault.get("faultstring", "Unknown error")
            elif "errors" in error_data:
                errors = error_data["errors"]
                if isinstance(errors, dict) and "error" in errors:
                    message = errors["error"]
                else:
                    message = str(errors)
            else:
                message = response.text
        except Exception:
            message = response.text or f"HTTP {response.status_code}"
        
        raise TradierAPIError(
            message=message,
            status_code=response.status_code,
        )


def validate_and_parse(response: Response) -> dict[str, Any]:
    """
    Validate an HTTP response and parse JSON data.
    
    :param response: The httpx Response object to validate and parse.
    :returns: The parsed JSON data as a dictionary.
    :raises TradierAPIError: If the response indicates an error.
    """
    validate_response(response)
    return response.json()  # type: ignore[no-any-return]


def parse_datetime(value: str | None) -> datetime | None:
    """
    Parse a datetime string from the Tradier API.
    
    :param value: The datetime string to parse.
    :returns: A datetime object or None if the input is None.
    """
    if value is None:
        return None
    
    # Tradier uses ISO 8601 format
    try:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    except ValueError:
        # Try alternative formats
        for fmt in ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%d"]:
            try:
                return datetime.strptime(value, fmt)
            except ValueError:
                continue
        raise ValueError(f"Unable to parse datetime: {value}") from None


def parse_date(value: str | None) -> date | None:
    """
    Parse a date string from the Tradier API.
    
    :param value: The date string to parse.
    :returns: A date object or None if the input is None.
    """
    if value is None:
        return None
    
    return datetime.strptime(value, "%Y-%m-%d").date()



from .candle import Candle
from .event import Event
from .greeks import Greeks
from .profile import Profile
from .quote import Quote
from .summary import Summary
from .theoprice import TheoPrice
from .timeandsale import TimeAndSale
from .trade import Trade
from .underlying import Underlying

__all__ = [
    "Candle",
    "Event",
    "Greeks",
    "Profile",
    "Quote",
    "Summary",
    "TheoPrice",
    "TimeAndSale",
    "Trade",
    "Underlying",
]



from datetime import date, datetime
from decimal import Decimal

from optrabot.broker.tastytrade.session import Session
from optrabot.broker.tastytrade.utils import TastytradeData, validate_and_parse


class DividendInfo(TastytradeData):
    """
    Dataclass representing dividend information for a given symbol.
    """

    occurred_date: date
    amount: Decimal


class EarningsInfo(TastytradeData):
    """
    Dataclass representing earnings information for a given symbol.
    """

    occurred_date: date
    eps: Decimal


class EarningsReport(TastytradeData):
    """
    Dataclass containing information about a recent earnings report, or the
    expected date of the next one.
    """

    estimated: bool
    late_flag: int
    visible: bool
    actual_eps: Decimal | None = None
    consensus_estimate: Decimal | None = None
    expected_report_date: date | None = None
    quarter_end_date: date | None = None
    time_of_day: str | None = None
    updated_at: datetime | None = None


class Liquidity(TastytradeData):
    """
    Dataclass representing liquidity information for a given symbol.
    """

    sum: Decimal
    count: int
    started_at: datetime
    updated_at: datetime | None = None


class OptionExpirationImpliedVolatility(TastytradeData):
    """
    Dataclass containing implied volatility information for a given symbol
    and expiration date.
    """

    expiration_date: date
    settlement_type: str
    option_chain_type: str
    implied_volatility: Decimal | None = None


class MarketMetricInfo(TastytradeData):
    """
    Dataclass representing market metrics for a given symbol.

    Contains lots of useful information, like IV rank, IV percentile and beta.
    """

    symbol: str
    implied_volatility_index: Decimal | None = None
    implied_volatility_index_5_day_change: Decimal | None = None
    implied_volatility_index_rank: str | None = None
    tos_implied_volatility_index_rank: Decimal | None = None
    tw_implied_volatility_index_rank: Decimal | None = None
    tos_implied_volatility_index_rank_updated_at: datetime | None = None
    implied_volatility_index_rank_source: str | None = None
    implied_volatility_percentile: str | None = None
    implied_volatility_updated_at: datetime | None = None
    liquidity_rating: int | None = None
    updated_at: datetime
    option_expiration_implied_volatilities: (
        list[OptionExpirationImpliedVolatility] | None
    ) = None  # noqa: E501
    beta: Decimal | None = None
    corr_spy_3month: Decimal | None = None
    market_cap: Decimal
    earnings: EarningsReport | None = None
    price_earnings_ratio: Decimal | None = None
    earnings_per_share: Decimal | None = None
    dividend_rate_per_share: Decimal | None = None
    implied_volatility_30_day: Decimal | None = None
    historical_volatility_30_day: Decimal | None = None
    historical_volatility_60_day: Decimal | None = None
    historical_volatility_90_day: Decimal | None = None
    iv_hv_30_day_difference: Decimal | None = None
    beta_updated_at: datetime | None = None
    created_at: datetime | None = None
    dividend_ex_date: date | None = None
    dividend_next_date: date | None = None
    dividend_pay_date: date | None = None
    dividend_updated_at: datetime | None = None
    liquidity_value: Decimal | None = None
    liquidity_rank: Decimal | None = None
    liquidity_running_state: Liquidity | None = None
    dividend_yield: Decimal | None = None
    listed_market: str | None = None
    lendability: str | None = None
    borrow_rate: Decimal | None = None


async def a_get_market_metrics(
    session: Session, symbols: list[str]
) -> list[MarketMetricInfo]:
    """
    Retrieves market metrics for the given symbols.

    :param session: active user session to use
    :param symbols: list of symbols to retrieve metrics for
    """
    data = await session._a_get(
        "/market-metrics", params={"symbols": ",".join(symbols)}
    )
    return [MarketMetricInfo(**i) for i in data["items"]]


def get_market_metrics(session: Session, symbols: list[str]) -> list[MarketMetricInfo]:
    """
    Retrieves market metrics for the given symbols.

    :param session: active user session to use
    :param symbols: list of symbols to retrieve metrics for
    """
    data = session._get("/market-metrics", params={"symbols": ",".join(symbols)})
    return [MarketMetricInfo(**i) for i in data["items"]]


async def a_get_dividends(session: Session, symbol: str) -> list[DividendInfo]:
    """
    Retrieves dividend information for the given symbol.

    :param session: active user session to use
    :param symbol: symbol to retrieve dividend information for
    """
    symbol = symbol.replace("/", "%2F")
    data = await session._a_get(
        f"/market-metrics/historic-corporate-events/dividends/{symbol}"
    )
    return [DividendInfo(**i) for i in data["items"]]


def get_dividends(session: Session, symbol: str) -> list[DividendInfo]:
    """
    Retrieves dividend information for the given symbol.

    :param session: active user session to use
    :param symbol: symbol to retrieve dividend information for
    """
    symbol = symbol.replace("/", "%2F")
    data = session._get(f"/market-metrics/historic-corporate-events/dividends/{symbol}")
    return [DividendInfo(**i) for i in data["items"]]


async def a_get_earnings(
    session: Session, symbol: str, start_date: date
) -> list[EarningsInfo]:
    """
    Retrieves earnings information for the given symbol.

    :param session: active user session to use
    :param symbol: symbol to retrieve earnings information for
    :param start_date: limits earnings to those on or after the given date
    """
    symbol = symbol.replace("/", "%2F")
    params = {"start-date": start_date}
    data = await session._a_get(
        (f"/market-metrics/historic-corporate-events/earnings-reports/{symbol}"),
        params=params,
    )
    return [EarningsInfo(**i) for i in data["items"]]


def get_earnings(session: Session, symbol: str, start_date: date) -> list[EarningsInfo]:
    """
    Retrieves earnings information for the given symbol.

    :param session: active user session to use
    :param symbol: symbol to retrieve earnings information for
    :param start_date: limits earnings to those on or after the given date
    """
    symbol = symbol.replace("/", "%2F")
    params = {"start-date": start_date}
    data = session._get(
        (f"/market-metrics/historic-corporate-events/earnings-reports/{symbol}"),
        params=params,
    )
    return [EarningsInfo(**i) for i in data["items"]]


async def a_get_risk_free_rate(session: Session) -> Decimal:
    """
    Retrieves the current risk-free rate.

    :param session: active user session to use
    """
    request = session.async_client.build_request(
        "GET", "/margin-requirements-public-configuration", timeout=30
    )
    del request.headers["Authorization"]
    response = await session.async_client.send(request)
    data = validate_and_parse(response)
    return Decimal(data["risk-free-rate"])


def get_risk_free_rate(session: Session) -> Decimal:
    """
    Retrieves the current risk-free rate.

    :param session: active user session to use
    """
    request = session.sync_client.build_request(
        "GET", "/margin-requirements-public-configuration", timeout=30
    )
    del request.headers["Authorization"]
    response = session.sync_client.send(request)
    data = validate_and_parse(response)
    return Decimal(data["risk-free-rate"])



from decimal import Decimal
from typing import Optional

from .event import Event


class Summary(Event):
    """
    Summary is an information snapshot about the trading session including
    session highs, lows, etc. This record has two goals: Transmit OHLC
    values, and provide data for charting. OHLC is required for a daily chart,
    and if an exchange does not provide it, the charting services refer to the
    Summary event.

    Before opening the bidding, the values are reset to N/A or NaN.
    """

    #: identifier of the day that this summary represents
    day_id: int
    #: the price type of the last (close) price for the day
    #: possible values are FINAL | INDICATIVE | PRELIMINARY | REGULAR
    day_close_price_type: str
    #: identifier of the previous day that this summary represents
    prev_day_id: int
    #: the price type of the last (close) price for the previous day
    #: possible values are FINAL | INDICATIVE | PRELIMINARY | REGULAR
    prev_day_close_price_type: str
    #: open interest of the symbol as the number of open contracts
    open_interest: int
    #: the first (open) price for the day
    day_open_price: Optional[Decimal] = None
    #: the maximal (high) price for the day
    day_high_price: Optional[Decimal] = None
    #: the minimal (low) price for the day
    day_low_price: Optional[Decimal] = None
    #: the last (close) price for the day
    day_close_price: Optional[Decimal] = None
    #: the last (close) price for the previous day
    prev_day_close_price: Optional[Decimal] = None
    #: total volume traded for the previous day
    prev_day_volume: Optional[Decimal] = None



from __future__ import annotations

import json
from datetime import date, datetime, timedelta
from typing import Any

from httpx import AsyncClient, Client
from typing_extensions import Self

from optrabot.broker.tastytrade import API_URL, API_VERSION, CERT_URL, logger
from optrabot.broker.tastytrade.utils import (
    TastytradeData,
    now_in_new_york,
    validate_and_parse,
    validate_response,
)


class Address(TastytradeData):
    """
    Dataclass containing customer address information.
    """

    city: str
    country: str
    is_domestic: bool
    is_foreign: bool
    postal_code: str
    state_region: str
    street_one: str
    street_two: str | None = None
    street_three: str | None = None


class EntityOfficer(TastytradeData):
    """
    Dataclass containing entity officer information.
    """

    id: str
    external_id: str
    first_name: str
    last_name: str
    middle_name: str
    prefix_name: str
    suffix_name: str
    address: Address
    birth_country: str
    birth_date: date
    citizenship_country: str
    email: str
    employer_name: str
    employment_status: str
    home_phone_number: str
    is_foreign: bool
    job_title: str
    marital_status: str
    mobile_phone_number: str
    number_of_dependents: int
    occupation: str
    owner_of_record: bool
    relationship_to_entity: str
    tax_number: str
    tax_number_type: str
    usa_citizenship_type: str
    visa_expiration_date: date
    visa_type: str
    work_phone_number: str


class EntitySuitability(TastytradeData):
    """
    Dataclass containing entity suitability information.
    """

    id: str
    annual_net_income: int
    covered_options_trading_experience: str
    entity_id: int
    futures_trading_experience: str
    liquid_net_worth: int
    net_worth: int
    stock_trading_experience: str
    tax_bracket: str
    uncovered_options_trading_experience: str


class CustomerAccountMarginType(TastytradeData):
    """
    Dataclass containing margin information for a customer account type.
    """

    name: str
    is_margin: bool


class CustomerAccountType(TastytradeData):
    """
    Dataclass containing information for a type of customer account.
    """

    name: str
    description: str
    is_tax_advantaged: bool
    is_publicly_available: bool
    has_multiple_owners: bool
    margin_types: list[CustomerAccountMarginType]


class CustomerEntity(TastytradeData):
    """
    Dataclass containing customer entity information.
    """

    id: str
    address: Address
    business_nature: str
    email: str
    entity_officers: list[EntityOfficer]
    entity_suitability: EntitySuitability
    entity_type: str
    foreign_institution: str
    grantor_birth_date: str
    grantor_email: str
    grantor_first_name: str
    grantor_last_name: str
    grantor_middle_name: str
    grantor_tax_number: str
    has_foreign_bank_affiliation: str
    has_foreign_institution_affiliation: str
    is_domestic: bool
    legal_name: str
    phone_number: str
    tax_number: str


class CustomerPerson(TastytradeData):
    """
    Dataclass containing customer person information.
    """

    external_id: str
    first_name: str
    last_name: str
    citizenship_country: str
    usa_citizenship_type: str
    employment_status: str
    marital_status: str
    number_of_dependents: int
    occupation: str | None = None
    middle_name: str | None = None
    prefix_name: str | None = None
    suffix_name: str | None = None
    birth_country: str | None = None
    birth_date: date | str | None = None
    visa_expiration_date: date | None = None
    visa_type: str | None = None
    employer_name: str | None = None
    job_title: str | None = None


class CustomerSuitability(TastytradeData):
    """
    Dataclass containing customer suitability information.
    """

    id: int
    annual_net_income: int
    covered_options_trading_experience: str
    employment_status: str
    futures_trading_experience: str
    liquid_net_worth: int
    marital_status: str
    net_worth: int
    number_of_dependents: int
    stock_trading_experience: str
    uncovered_options_trading_experience: str
    customer_id: str | None = None
    employer_name: str | None = None
    job_title: str | None = None
    occupation: str | None = None
    tax_bracket: str | None = None


class Customer(TastytradeData):
    """
    Dataclass containing customer information.
    """

    id: str
    first_name: str
    first_surname: str
    last_name: str
    address: Address
    customer_suitability: CustomerSuitability
    mailing_address: Address
    is_foreign: bool
    regulatory_domain: str
    usa_citizenship_type: str
    home_phone_number: str
    mobile_phone_number: str
    work_phone_number: str
    birth_date: date
    email: str
    external_id: str
    tax_number: str
    tax_number_type: str
    citizenship_country: str
    agreed_to_margining: bool
    subject_to_tax_withholding: bool
    agreed_to_terms: bool
    ext_crm_id: str
    has_industry_affiliation: bool
    has_listed_affiliation: bool
    has_political_affiliation: bool
    has_delayed_quotes: bool
    has_pending_or_approved_application: bool
    is_professional: bool
    permitted_account_types: list[CustomerAccountType]
    created_at: datetime
    identifiable_type: str
    person: CustomerPerson
    gender: str | None = None
    middle_name: str | None = None
    prefix_name: str | None = None
    second_surname: str | None = None
    suffix_name: str | None = None
    foreign_tax_number: str | None = None
    birth_country: str | None = None
    visa_expiration_date: date | None = None
    visa_type: str | None = None
    signature_of_agreement: bool | None = None
    desk_customer_id: str | None = None
    entity: CustomerEntity | None = None
    family_member_names: str | None = None
    has_institutional_assets: str | bool | None = None
    industry_affiliation_firm: str | None = None
    is_investment_adviser: bool | None = None
    listed_affiliation_symbol: str | None = None
    political_organization: str | None = None
    user_id: str | None = None


_fmt = "%Y-%m-%d %H:%M:%S%z"


class Session:
    """
    Contains a managed user login which can then be used to interact with the
    remote API.

    :param provider_secret: OAuth secret for your provider
    :param refresh_token: refresh token for the user
    :param is_test:
        whether to use the test API endpoints, default False
    :param proxy:
        if provided, all requests will be made through this proxy, as well as
        web socket connections for streamers.
    """

    def __init__(
        self,
        provider_secret: str,
        refresh_token: str,
        is_test: bool = False,
        proxy: str | None = None,
    ):
        #: Whether this is a cert or real session
        self.is_test = is_test
        #: Proxy URL to use for requests and web sockets
        self.proxy = proxy
        #: OAuth secret for your provider
        self.provider_secret = provider_secret
        #: Refresh token for the user
        self.refresh_token = refresh_token
        # The headers to use for API requests
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if not is_test:  # not accepted in sandbox
            headers["Accept-Version"] = API_VERSION
        #: httpx client for sync requests
        self.sync_client = Client(
            base_url=(CERT_URL if is_test else API_URL), headers=headers, proxy=proxy
        )
        #: httpx client for async requests
        self.async_client = AsyncClient(
            base_url=self.sync_client.base_url, headers=headers, proxy=proxy
        )
        #: expiration for streamer token
        self.streamer_expiration = now_in_new_york()
        self.refresh()

    def _streamer_refresh(self, data: Any) -> None:
        # Auth token for dxfeed websocket
        self.streamer_token = data["token"]
        # URL for dxfeed websocket
        self.dxlink_url = data["dxlink-url"]
        self.streamer_expiration = datetime.fromisoformat(
            data["expires-at"].replace("Z", "+00:00")
        )

    def refresh(self) -> None:
        """
        Refreshes the acccess token using the stored refresh token.
        Also refreshes the streamer token if necessary.
        """
        request = self.sync_client.build_request(
            "POST",
            "/oauth/token",
            json={
                "grant_type": "refresh_token",
                "client_secret": self.provider_secret,
                "refresh_token": self.refresh_token,
            },
        )
        # Don't send the Authorization header for this request
        request.headers.pop("Authorization", None)
        response = self.sync_client.send(request)
        validate_response(response)
        data = response.json()
        #: The session token used to authenticate requests
        self.session_token = data["access_token"]
        token_lifetime: int = data.get("expires_in", 900)
        #: expiration for session token
        self.session_expiration = now_in_new_york() + timedelta(seconds=token_lifetime)
        logger.debug(f"Refreshed token, expires in {token_lifetime}ms")
        auth_headers = {"Authorization": f"Bearer {self.session_token}"}
        # update the httpx clients with the new token
        self.sync_client.headers.update(auth_headers)
        self.async_client.headers.update(auth_headers)
        # update the streamer token if necessary
        if not self.is_test and self.streamer_expiration < self.session_expiration:
            data = self._get("/api-quote-tokens")
            self._streamer_refresh(data)

    async def a_refresh(self) -> None:
        """
        Refreshes the acccess token using the stored refresh token.
        Also refreshes the streamer token if necessary.
        """
        request = self.async_client.build_request(
            "POST",
            "/oauth/token",
            json={
                "grant_type": "refresh_token",
                "client_secret": self.provider_secret,
                "refresh_token": self.refresh_token,
            },
        )
        # Don't send the Authorization header for this request
        request.headers.pop("Authorization", None)
        response = await self.async_client.send(request)
        validate_response(response)
        data = response.json()
        # update the relevant tokens
        self.session_token = data["access_token"]
        token_lifetime: int = data.get("expires_in", 900)
        self.session_expiration = now_in_new_york() + timedelta(token_lifetime)
        logger.debug(f"Refreshed token, expires in {token_lifetime}ms")
        auth_headers = {"Authorization": f"Bearer {self.session_token}"}
        # update the httpx clients with the new token
        self.sync_client.headers.update(auth_headers)
        self.async_client.headers.update(auth_headers)
        # update the streamer token if necessary
        if not self.is_test and self.streamer_expiration < self.session_expiration:
            # Pull streamer tokens and urls
            data = await self._a_get("/api-quote-tokens")
            self._streamer_refresh(data)

    async def a_get_customer(self) -> Customer:
        """
        Gets the customer dict from the API.
        """
        data = await self._a_get("/customers/me")
        return Customer(**data)

    def get_customer(self) -> Customer:
        """
        Gets the customer dict from the API.
        """
        data = self._get("/customers/me")
        return Customer(**data)

    async def a_validate(self) -> bool:
        """
        Validates the current session by sending a request to the API.
        """
        response = await self.async_client.post("/sessions/validate")
        return response.status_code // 100 == 2

    def validate(self) -> bool:
        """
        Validates the current session by sending a request to the API.
        """
        response = self.sync_client.post("/sessions/validate")
        return response.status_code // 100 == 2

    def serialize(self) -> str:
        """
        Serializes the session to a string, useful for storing
        a session for later use.
        Could be used with pickle, Redis, etc.
        """
        attrs = self.__dict__.copy()
        del attrs["async_client"]
        del attrs["sync_client"]
        attrs["session_expiration"] = self.session_expiration.strftime(_fmt)
        attrs["streamer_expiration"] = self.streamer_expiration.strftime(_fmt)
        attrs["headers"] = dict(self.sync_client.headers.copy())
        return json.dumps(attrs)

    @classmethod
    def deserialize(cls, serialized: str) -> Self:
        """
        Create a new Session object from a serialized string.
        """
        deserialized: dict[str, Any] = json.loads(serialized)
        headers = deserialized.pop("headers")
        self = cls.__new__(cls)
        self.__dict__ = deserialized
        base_url = CERT_URL if self.is_test else API_URL
        self.session_expiration = datetime.strptime(
            deserialized["session_expiration"], _fmt
        )
        self.streamer_expiration = datetime.strptime(
            deserialized["streamer_expiration"], _fmt
        )
        self.sync_client = Client(base_url=base_url, headers=headers, proxy=self.proxy)
        self.async_client = AsyncClient(
            base_url=base_url, headers=headers, proxy=self.proxy
        )
        return self

    async def _a_get(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = await self.async_client.get(url, timeout=30, **kwargs)
        return validate_and_parse(response)

    def _get(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = self.sync_client.get(url, timeout=30, **kwargs)
        return validate_and_parse(response)

    async def _a_delete(self, url: str, **kwargs: Any) -> None:
        response = await self.async_client.delete(url, **kwargs)
        validate_response(response)

    def _delete(self, url: str, **kwargs: Any) -> None:
        response = self.sync_client.delete(url, **kwargs)
        validate_response(response)

    async def _a_post(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = await self.async_client.post(url, **kwargs)
        return validate_and_parse(response)

    def _post(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = self.sync_client.post(url, **kwargs)
        return validate_and_parse(response)

    async def _a_put(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = await self.async_client.put(url, **kwargs)
        return validate_and_parse(response)

    def _put(self, url: str, **kwargs: Any) -> dict[str, Any]:
        response = self.sync_client.put(url, **kwargs)
        return validate_and_parse(response)



import logging

API_URL = "https://api.tastyworks.com"
API_VERSION = "20251101"
CERT_URL = "https://api.cert.tastyworks.com"
VERSION = "11.1.0"

__version__ = VERSION
version_str: str = f"tastyware/tastytrade:v{VERSION}"

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# ruff: noqa: E402

from .account import Account
from .session import Session
from .streamer import AlertStreamer, DXLinkStreamer

__all__ = ["Account", "AlertStreamer", "DXLinkStreamer", "Session"]



from decimal import Decimal
from typing import Optional

from .event import Event


class Quote(Event):
    """
    A Quote event is a snapshot of the best bid and ask prices, and other
    fields that change with each quote.
    """

    #: sequence of this quote
    sequence: int
    #: microseconds and nanoseconds part of time of the last bid or ask change
    time_nano_part: int
    #: time of the last bid change
    bid_time: int
    #: bid exchange code
    bid_exchange_code: str
    #: time of the last ask change
    ask_time: int
    #: ask exchange code
    ask_exchange_code: str
    #: bid price
    bid_price: Decimal
    #: ask price
    ask_price: Decimal
    #: bid size as integer number (rounded toward zero)
    #: or decimal for cryptocurrencies
    bid_size: Optional[Decimal] = None
    #: ask size as integer number (rounded toward zero)
    #: or decimal for cryptocurrencies
    ask_size: Optional[Decimal] = None



from datetime import date, datetime, timedelta
from decimal import Decimal
from enum import Enum
from json import JSONDecodeError
from typing import Any, Type, TypeVar, cast
from zoneinfo import ZoneInfo

from httpx import AsyncClient, Client, Response
from pandas import Timestamp
from pandas_market_calendars import get_calendar  # type: ignore[import-untyped]
from pydantic import BaseModel, ConfigDict

from optrabot.broker.tastytrade import logger

NYSE: Any = get_calendar("NYSE")
TZ = ZoneInfo("US/Eastern")


class PriceEffect(str, Enum):
    """
    This is an :class:`~enum.Enum` that shows the sign of a price effect, since
    Tastytrade is apparently against negative numbers.
    """

    CREDIT = "Credit"
    DEBIT = "Debit"
    NONE = "None"


def now_in_new_york() -> datetime:
    """
    Gets the current time in the New York timezone.
    """
    return datetime.now(TZ)


def today_in_new_york() -> date:
    """
    Gets the current date in the New York timezone.
    """
    return now_in_new_york().date()


def is_market_open_now() -> bool:
    """
    Check if the market is currently open.
    """
    today = today_in_new_york()
    sched = NYSE.schedule(start_date=today, end_date=today)
    if sched.empty:
        # Closed full day (weekend or holiday)
        return False

    # Use iloc[0] since schedule has only one row for a single day
    market_open: Timestamp = sched.iloc[0]["market_open"]
    market_close: Timestamp = sched.iloc[0]["market_close"]
    return market_open <= now_in_new_york() < market_close


def is_market_open_on(day: date | None = None) -> bool:
    """
    Returns whether the market was/is/will be open at ANY point
    during the given day.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = day or today_in_new_york()
    date_range = NYSE.valid_days(day, day)
    return not date_range.empty


def get_third_friday(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the month of the given date,
    or the monthly expiration associated with today's month.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = (day or today_in_new_york()).replace(day=1) + timedelta(weeks=2)
    while day.weekday() != 4:  # Friday
        day += timedelta(days=1)
    return day


def get_tasty_monthly() -> date:
    """
    Gets the monthly expiration closest to 45 days from the current date.
    """
    day = today_in_new_york()
    exp1 = get_third_friday(day + timedelta(weeks=4))
    exp2 = get_third_friday(day + timedelta(weeks=8))
    day45 = day + timedelta(days=45)
    return exp1 if day45 - exp1 < exp2 - day45 else exp2


def _get_last_day_of_month(day: date) -> date:
    if day.month == 12:
        last = day.replace(day=1, month=1, year=day.year + 1)
    else:
        last = day.replace(day=1, month=day.month + 1)
    return last - timedelta(days=1)


def get_future_fx_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the FX futures: /6E, /6A, etc.
    As far as I can tell, these expire on the first Friday prior to the second
    Wednesday.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = (day or today_in_new_york()).replace(day=1) + timedelta(weeks=1)
    while day.weekday() != 2:  # Wednesday
        day += timedelta(days=1)
    while day.weekday() != 4:  # Friday
        day -= timedelta(days=1)
    return day


def get_future_treasury_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the treasury futures: /ZN,
    /ZB, etc. According to CME, these expire the Friday before the 2nd last
    business day of the month. If this is not a business day, they expire 1
    business day prior.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = day or today_in_new_york()
    last_day = _get_last_day_of_month(day)
    first_day = last_day.replace(day=1)
    valid_range: list[date] = [d.date() for d in NYSE.valid_days(first_day, last_day)]
    itr = valid_range[-2] - timedelta(days=1)
    while itr.weekday() != 4:  # Friday
        itr -= timedelta(days=1)
    if itr in valid_range:
        return itr
    return itr - timedelta(days=1)


def get_future_metal_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the metals futures: /GC, /SI,
    etc. According to CME, these expire on the 4th last business day of the
    month, unless that day occurs on a Friday or the day before a holiday, in
    which case they expire on the prior business day.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = day or today_in_new_york()
    last_day = _get_last_day_of_month(day)
    first_day = last_day.replace(day=1)
    valid_range: list[date] = [d.date() for d in NYSE.valid_days(first_day, last_day)]
    itr = valid_range[-4]
    next_day = itr + timedelta(days=1)
    if itr.weekday() == 4 or next_day not in valid_range:
        return valid_range[-5]
    return itr


def get_future_grain_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the grain futures: /ZC, /ZW,
    etc. According to CME, these expire on the Friday which precedes, by at
    least 2 business days, the last business day of the month.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = day or today_in_new_york()
    last_day = _get_last_day_of_month(day)
    first_day = last_day.replace(day=1)
    valid_range: list[date] = [d.date() for d in NYSE.valid_days(first_day, last_day)]
    itr = valid_range[-3]
    while itr.weekday() != 4:  # Friday
        itr -= timedelta(days=1)
    return itr


def get_future_oil_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the WTI oil futures: /CL and
    /MCL. According to CME, these expire 6 business days before the 25th day
    of the month, unless the 25th day is not a business day, in which case
    they expire 7 business days prior to the 25th day of the month.

    :param day: date to check. If not provided defaults to current NY date.
    """
    last_day = (day or today_in_new_york()).replace(day=25)
    first_day = last_day.replace(day=1)
    valid_range: list[date] = [d.date() for d in NYSE.valid_days(first_day, last_day)]
    return valid_range[-7]


def get_future_index_monthly(day: date | None = None) -> date:
    """
    Gets the monthly expiration associated with the index futures: /ES, /RTY,
    /NQ, etc. According to CME, these expire on the last business day of the
    month.

    :param day: date to check. If not provided defaults to current NY date.
    """
    day = day or today_in_new_york()
    last_day = _get_last_day_of_month(day)
    first_day = last_day.replace(day=1)
    valid_range: list[date] = [d.date() for d in NYSE.valid_days(first_day, last_day)]
    return valid_range[-1]


class TastytradeError(Exception):
    """
    An internal error raised by the Tastytrade SDK.
    """

    pass


def _dasherize(s: str) -> str:
    """
    Converts a string from snake case to dasherized.

    :param s: string to convert
    """
    return s.replace("_", "-")


class TastytradeData(BaseModel):
    """
    A pydantic dataclass that converts keys from snake case to dasherized
    and performs type validation and coercion.
    """

    model_config = ConfigDict(alias_generator=_dasherize, populate_by_name=True)

    def __str__(self) -> str:
        return " ".join(f"{a}={v!r}" for a, v in self.__repr_args__() if v)

    def __repr__(self) -> str:
        return f"{self.__repr_name__()}({str(self)})"  # type: ignore


def validate_response(response: Response) -> None:
    """
    Checks if the given code is an error; if so, raises an exception.

    :param response: response to check for errors
    """
    if response.status_code // 100 != 2:
        try:
            json: dict[str, Any] = response.json()
        except JSONDecodeError as e:
            raise TastytradeError(f"Couldn't parse response: {response.text}") from e
        if not (content := json.get("error")):
            raise TastytradeError(f"Couldn't parse response: {json}")
        errors = content.get("errors") or [content]
        message = ""
        for error in errors:
            if "code" in error and "message" in error:
                message += f"{error['code']}: {error['message']}\n"
            elif "domain" in error and "reason" in error:
                message += f"{error['domain']}: {error['reason']}\n"
            else:
                logger.debug(f"Unknown error type: {error}")

        raise TastytradeError(message)


def validate_and_parse(response: Response) -> dict[str, Any]:
    """
    Checks if the given code is an error; if so, raises an exception.
    Then, returns the JSON payload.

    :param response: response to check for errors
    """
    validate_response(response)
    json = response.json()
    if not (data := json.get("data")):
        raise TastytradeError(f"No data present in response: {json}")
    return cast(dict[str, Any], data)


def get_sign(value: Decimal | None) -> PriceEffect | None:
    """
    Get a PriceEffect for a signed value.

    :param value: value to check
    """
    if not value:
        return None
    return PriceEffect.DEBIT if value < 0 else PriceEffect.CREDIT


def set_sign_for(data: Any, properties: list[str]) -> Any:
    """
    Handles setting the sign of a number using the associated "-effect" field.

    :param data: the raw, unprocessed model object
    :param properties: the name of the number fields to set
    """
    if isinstance(data, dict):
        data = cast(dict[str, Any], data)
        for property in properties:
            key = _dasherize(property)
            if data.get(f"{key}-effect") == PriceEffect.DEBIT:
                data[key] = -abs(Decimal(data[key]))
    return data


T = TypeVar("T", bound=TastytradeData)


def paginate(
    client: Client, model: Type[T], url: str, params: dict[str, Any]
) -> list[T]:
    """
    Helper for paginated endpoints. Excepts params to have at least `page-offset` and
    `per-page` parameters.
    If `params["page-offset"]` is None, iterates over all results; otherwise, gets a
    specific page.

    :param client: the httpx client for making request
    :param model: the TastytradeData model for results
    :param url: the endpoint to use
    :param params: parameters for request
    """
    res: list[T] = []
    # if a specific page is provided, we just get that page;
    # otherwise, we loop through all pages
    paginate = False
    if params["page-offset"] is None:
        params["page-offset"] = 0
        paginate = True
    params = {k: v for k, v in params.items() if v is not None}
    # loop through pages and get all transactions
    while True:
        response = client.get(url, params=params)
        validate_response(response)
        json = response.json()
        res.extend([model(**i) for i in json["data"]["items"]])
        # handle pagination
        pagination = json.get("pagination")
        if (
            not pagination
            or not paginate
            or pagination["page-offset"] >= pagination["total-pages"] - 1
        ):
            break
        params["page-offset"] += 1
    return res


async def a_paginate(
    client: AsyncClient, model: Type[T], url: str, params: dict[str, Any]
) -> list[T]:
    """
    Helper for paginated endpoints. Excepts params to have at least `page-offset` and
    `per-page` parameters.
    If `params["page-offset"]` is None, iterates over all results; otherwise, gets a
    specific page.

    :param client: the httpx client for making request
    :param model: the TastytradeData model for results
    :param url: the endpoint to use
    :param params: parameters for request
    """
    res: list[T] = []
    # if a specific page is provided, we just get that page;
    # otherwise, we loop through all pages
    paginate = False
    if params["page-offset"] is None:
        params["page-offset"] = 0
        paginate = True
    params = {k: v for k, v in params.items() if v is not None}
    # loop through pages and get all transactions
    while True:
        response = await client.get(url, params=params)
        validate_response(response)
        json = response.json()
        res.extend([model(**i) for i in json["data"]["items"]])
        # handle pagination
        pagination = json.get("pagination")
        if (
            not pagination
            or not paginate
            or pagination["page-offset"] >= pagination["total-pages"] - 1
        ):
            break
        params["page-offset"] += 1
    return res



"""
Custom exceptions for the Tradier API client.
"""

from __future__ import annotations


class TradierError(Exception):
    """
    Base exception for all Tradier API errors.
    """

    pass


class TradierAPIError(TradierError):
    """
    Exception raised when the Tradier API returns an error response.
    
    :param message: The error message from the API.
    :param status_code: The HTTP status code of the response.
    """

    def __init__(self, message: str, status_code: int | None = None) -> None:
        self.message = message
        self.status_code = status_code
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        if self.status_code:
            return f"[{self.status_code}] {self.message}"
        return self.message

    def __repr__(self) -> str:
        return f"TradierAPIError(message={self.message!r}, status_code={self.status_code!r})"


class TradierAuthenticationError(TradierAPIError):
    """
    Exception raised when authentication fails.
    """

    def __init__(self, message: str = "Authentication failed") -> None:
        super().__init__(message=message, status_code=401)


class TradierRateLimitError(TradierAPIError):
    """
    Exception raised when the API rate limit is exceeded.
    """

    def __init__(self, message: str = "Rate limit exceeded") -> None:
        super().__init__(message=message, status_code=429)


class TradierNotFoundError(TradierAPIError):
    """
    Exception raised when a requested resource is not found.
    """

    def __init__(self, message: str = "Resource not found") -> None:
        super().__init__(message=message, status_code=404)
